@@ -3,6 +3,8 @@ from typing import Dict, Tuple, Optional, Any
 import boto3
 import botocore.exceptions
 import boto3.dynamodb.conditions
+import circuitbreaker
+import concurrent.futures
 
 
 class ItemKey:
@@ -21,10 +23,18 @@ class ItemKey:
         return {*{self.name: self.value}, *(self.secondary.as_dynamodb_key() if self.secondary else {})}
 
 
+class InternalError(Exception):
+    pass
+
+
 class DB:
 
-    def __init__(self, name: str):
+    def __init__(self, name: str, logger):
         self._table = boto3.resource("dynamodb").Table(name)
+        self._logger = logger
+        self._to_put = []
+        self._to_update = []
+        self._to_delete = []
 
     def query_items(self, key: ItemKey):
         # query primary key
@@ -42,9 +52,76 @@ class DB:
             resp = self._table.get_item(Key=key.as_dynamodb_key())
             return resp["Item"]
         except botocore.exceptions.ClientError as e:
-            raise
+            raise InternalError() from e
+
+    def stage_put(self, key: ItemKey, item: Dict, expect_if_item_exists: Dict = None):
+        self._to_put.append((key, item, expect_if_item_exists))
+
+    def stage_update(self, key: ItemKey, item: Dict, old_item: Dict):
+        self._to_update.append((key, item, old_item))
+
+    def stage_delete(self, key: ItemKey):
+        self._to_delete.append((key,))
+
+    def flush(self):
+        # TODO make use of self._table.batch_writer for put/ delete
+        with concurrent.futures.ThreadPoolExecutor(max_workers=16) as executor:
+            for to_put in self._to_put:
+                executor.submit(self.put, *to_put)
+            for to_update in self._to_update:
+                executor.submit(self.update, *to_update)
+            for to_delete in self._to_delete:
+                executor.submit(self.delete, *to_delete)
 
     def put(self, key: ItemKey, item: Dict, expect_if_item_exists: Dict = None):
+        try:
+            self._operation("put", (key, item, expect_if_item_exists))
+        except InternalError as e:
+            if self._logger:
+                self._logger.exception(f"Put operation failed: {e}")
+            raise
+        except circuitbreaker.CircuitBreakerError as e:
+            if self._logger:
+                self._logger.exception(f"Didn't execute the operation because previous attempts failed")
+            raise InternalError() from e
+
+    def update(self, key: ItemKey, item: Dict, old_item: Dict):
+        try:
+            self._operation("update", (key, item, old_item))
+        except InternalError as e:
+            if self._logger:
+                self._logger.exception(f"Update operation failed: {e}")
+            raise
+        except circuitbreaker.CircuitBreakerError as e:
+            if self._logger:
+                self._logger.exception(f"Didn't execute the operation because previous attempts failed")
+            raise InternalError() from e
+
+    def delete(self, key: ItemKey):
+        try:
+            self._operation("delete", (key,))
+        except InternalError as e:
+            if self._logger:
+                self._logger.exception(f"Delete operation failed: {e}")
+            raise
+        except circuitbreaker.CircuitBreakerError as e:
+            if self._logger:
+                self._logger.exception(f"Didn't execute the operation because previous attempts failed")
+            raise InternalError() from e
+
+    @circuitbreaker.circuit(failure_threshold=1, recovery_timeout=None, expected_exception=InternalError)
+    def _operation(self, name: str, args):
+        # Aggregate operations for circuit breaker
+        if name == "put":
+            self._put(*args)
+        elif name == "update":
+            self._update(*args)
+        elif name == "delete":
+            self._delete(*args)
+        else:
+            raise ValueError("Invalid operation")
+
+    def _put(self, key: ItemKey, item: Dict, expect_if_item_exists: Dict = None):
         assert key.name in item
         assert not key.secondary or key.secondary.name in item
         expression_values = {}
@@ -63,7 +140,7 @@ class DB:
                 raise ExpectationNotMet() from e
             raise InternalError() from e
 
-    def update(self, key: ItemKey, item: Dict, old_item: Dict):
+    def _update(self, key: ItemKey, item: Dict, old_item: Dict):
         assert key.name in item
         assert not key.secondary or key.secondary.name in item
 
@@ -105,7 +182,7 @@ class DB:
         except botocore.exceptions.ClientError as e:
             raise InternalError() from e
 
-    def delete(self, key: ItemKey):
+    def _delete(self, key: ItemKey):
         try:
             self._table.delete_item(Key=key.as_dynamodb_key())
         except botocore.exceptions.ClientError as e:
@@ -115,7 +192,9 @@ class DB:
 class DBMock:
 
     def __init__(self, *args, **kwargs):
-        pass
+        self._to_put = []
+        self._to_update = []
+        self._to_delete = []
 
     _table: Dict[Tuple[str, Optional[str]], Any]
 
@@ -136,6 +215,24 @@ class DBMock:
             DBMock.count_get_operations += 1
         return DBMock._table.get(self._key_for_table(key))
 
+    def stage_put(self, key: ItemKey, item: Dict, expect_if_item_exists: Dict = None):
+        self._to_put.append((key, item, expect_if_item_exists))
+
+    def stage_update(self, key: ItemKey, item: Dict, old_item: Dict):
+        self._to_update.append((key, item, old_item))
+
+    def stage_delete(self, key: ItemKey):
+        self._to_delete.append((key,))
+
+    def flush(self):
+        with concurrent.futures.ThreadPoolExecutor(max_workers=16) as executor:
+            for to_put in self._to_put:
+                executor.submit(self.put, *to_put)
+            for to_update in self._to_update:
+                executor.submit(self.update, *to_update)
+            for to_delete in self._to_delete:
+                executor.submit(self.delete, *to_delete)
+
     count_put_operations: int
 
     def put(self, key: ItemKey, item: Dict, expect_if_item_exists: Dict = None, count=True):
@@ -182,7 +279,3 @@ class DBMock:
 
 class ExpectationNotMet(ValueError):
     pass
-
-
-class InternalError(Exception):
-    pass
